use std::cmp::max;

pub struct WithLengthAndWidth<T> {
    length: usize,
    width: usize,
    value: T,
}

pub enum RenderedCommon {
    Empty,
    Resolution,
    Event(String),
    Action {
        name: String,
        rendered: Box<WithLengthAndWidth<RenderedAny>>,
    },
}

pub enum RenderedWide {
    Common(RenderedCommon),
    Wide(Vec<WithLengthAndWidth<RenderedLong>>),
}

pub enum RenderedLong {
    Common(RenderedCommon),
    Long(Vec<WithLengthAndWidth<RenderedWide>>),
}

pub enum RenderedAny {
    Common(RenderedCommon),
    Wide(Vec<WithLengthAndWidth<RenderedLong>>),
    Long(Vec<WithLengthAndWidth<RenderedWide>>),
}

impl WithLengthAndWidth<RenderedWide> {
    pub fn push_into(self, vec: &mut WithLengthAndWidth<Vec<WithLengthAndWidth<RenderedWide>>>) {
        vec.length += self.length;
        vec.width = max(vec.width, self.width);
        vec.value.push(self);
    }
}

impl WithLengthAndWidth<RenderedLong> {
    pub fn push_into(self, vec: &mut WithLengthAndWidth<Vec<WithLengthAndWidth<RenderedLong>>>) {
        vec.width += self.width;
        vec.length = max(vec.length, self.length);
        vec.value.push(self);
    }
}

impl<A> WithLengthAndWidth<A> {
    pub fn map<B>(self, f: impl FnOnce(A) -> B) -> WithLengthAndWidth<B> {
        WithLengthAndWidth {
            length: self.length,
            width: self.width,
            value: f(self.value),
        }
    }

    pub fn length(&self) -> usize {
        self.length
    }

    pub fn width(&self) -> usize {
        self.width
    }

    pub fn value(&self) -> &A {
        &self.value
    }
}

impl<A: Default> Default for WithLengthAndWidth<A> {
    fn default() -> Self {
        Self {
            length: 0,
            width: 0,
            value: Default::default(),
        }
    }
}

impl WithLengthAndWidth<RenderedAny> {
    pub fn wrap(self, name: &str) -> WithLengthAndWidth<RenderedCommon> {
        WithLengthAndWidth {
            length: self.length,
            width: self.width,
            value: RenderedCommon::Action {
                name: name.into(),
                rendered: self.into(),
            },
        }
    }
}

impl RenderedCommon {
    pub fn empty() -> WithLengthAndWidth<Self> {
        WithLengthAndWidth {
            length: 0,
            width: 0,
            value: RenderedCommon::Empty,
        }
    }

    pub fn resolution() -> WithLengthAndWidth<Self> {
        WithLengthAndWidth {
            length: 1,
            width: 1,
            value: RenderedCommon::Resolution,
        }
    }

    pub fn event(event: &str) -> WithLengthAndWidth<Self> {
        WithLengthAndWidth {
            length: 0,
            width: 0,
            value: RenderedCommon::Event(event.into()),
        }
    }
}
